
#
# mapmaker - various tools to transform data
#

from datafile import dictlist_from_csv_str
from datamap import dynobj_from_dict
import ciso8601
import time


def object_convert_with_map(map: str, inobj: object) -> object:
    retdict = {}
    mapping = dictlist_from_csv_str(map)
    for field in mapping:
        toname = field.get("tofield", None)
        action = field.get("convert", None)
        if toname is None or action is None:
            continue
        if action == "" or action == "ignore":
            retdict[toname] = ""
        elif action == "copy" or action == "copy-unixtime":
            fromname = field.get("fromfield", None)
            if fromname is None:
                raise KeyError
            newval = getattr(inobj, fromname, None)
            if newval is None:
                newval = ""
            else:
                if action == "copy-unixtime":
                    ts = ciso8601.parse_datetime(newval)
                    # to get time in seconds:
                    newval = time.mktime(ts.timetuple())
            retdict[toname] = newval
        elif action == "copyif":
            fromnames1 = field.get("fromfield", None)
            if fromnames1 is None:
                raise KeyError
            retdict[toname] = ""
            fromnames = fromnames1.split("-")
            for fromname in fromnames:
                newval = getattr(inobj, fromname, None)
                if newval is not None:
                    retdict[toname] = newval
                    break
        elif action == "copyif-unixtime":
            fromnames1 = field.get("fromfield", None)
            if fromnames1 is None:
                raise KeyError
            retdict[toname] = ""
            fromnames = fromnames1.split("-")
            for fromname in fromnames:
                newval = getattr(inobj, fromname, None)
                if newval is not None:
                    ts = ciso8601.parse_datetime(newval)
                    # to get time in seconds:
                    newval = time.mktime(ts.timetuple())
                    retdict[toname] = newval
                    break
        elif action == "set":
            newval = field.get("value", "")
            retdict[toname] = newval
        else:
            raise KeyError("Invalid action {}".format(action))
    retobj = dynobj_from_dict(retdict, True)
    return retobj


def object_empty_with_map(map: str) -> object:
    retdict = {}
    mapping = dictlist_from_csv_str(map)
    for field in mapping:
        toname = field.get("tofield", None)
        retdict[toname] = None
    retobj = dynobj_from_dict(retdict, True)
    return retobj


#
# datafile - A quick library of data file parsing.
#

from .fileset import FileSet
from .fileparse import dict_from_urlfile
from .fileparse import dict_from_kvpfile
from .fileparse import dicts_from_files
from .fileparse import data_from_json_file
from .fileparse import dictlist_from_csv_file
from .fileparse import dictlist_from_csv_str
from .fileparse import dictlist_from_csv_stream
from .fileparse import column_from_csv_str
from .fileparse import column_from_csv_file
from .fileparse import data_to_json_file




#
# fileset - Operations on a collection of files.
#

#from . import ipmask
import glob
import os

class FileSet:
    def __init__(self):
        self.filepaths= []
        self.dirpath = ""
        self.globstr = "*.*"
        self.fullpath = ""
        self.iteridx = -1

    def glob(self, dirpath: str, globstr="*.*") -> int:
        self.dirpath = dirpath
        self.globstr = globstr
        if not os.path.isdir(self.dirpath):
            return None
        self.fullpath = "{}/{}".format(self.dirpath, self.globstr)
        fileiter = glob.iglob(self.fullpath)
        for path in fileiter:
            self.filepaths.append(path)
        return len(self.filepaths)

    def clear(self):
        self.filepaths = []

    def append(self, fp):
        self.filepaths.append(fp)

    def pop(self, idx: int) -> str:
        self.filepaths.pop(idx)

    def __iter__(self) -> object:
        self.iteridx = -1
        return self

    def __next__(self) -> str:
        if (self.iteridx + 1) >= len(self.filepaths):
            raise StopIteration
        self.iteridx += 1
        return self.filepaths[self.iteridx]







#
# objutils - Various object utilities.
#

from collections import namedtuple
from typing import Union
from types import SimpleNamespace
from copy import copy
import inspect

def ikeyval(inkey):
    keyval = inkey
    if is_number(inkey) and inkey > 1000000000:
        keyval = str(inkey)
    return keyval

def fixedobj_from_dict(aname: str, indict: dict, normalize=True) -> object:
    adict = copy(indict)
    if normalize:
        for key, val in adict.items():
            adict[ikeyval(key)] = normalize_value(val)
    retobj  = namedtuple(aname, adict.keys())(*adict.values())
    return retobj


def dict_from_object(anobj: object, fieldlist=None, missing=None) -> dict:
    retdict = {}
    if fieldlist is None:
        fieldlist = anobj.__dict__.keys()
    for fname in fieldlist:
        fval = getattr(anobj, fname, missing)
        retdict[fname] = fval
    return retdict


def dynobj_from_dict(indict: dict, normalize=True) -> object:
    adict = copy(indict)
    if normalize:
        for key, val in adict.items():
            adict[ikeyval(key)] = normalize_value(val)
    retobj  = SimpleNamespace()
    for key, val in adict.items():
        setattr(retobj, ikeyval(key), val)
    return retobj


def objlist_from_dictlist(aname: str, alist: list) -> list:
    retlist = []
    for ditem in alist:
        oitem = dynobj_from_dict(ditem)
        retlist.append(oitem)
    return retlist


def getvalue(anitem: Union[dict,object], keyprop:str):
    if inspect.isclass(type(anitem)):
        keyval = getattr(anitem, keyprop, None)
    elif type(anitem) == dict:
        if keyprop == dict:
            keyval = anitem[keyprop]
        else:
            keyval = None
    else:
        raise TypeError
    return keyval


def dict_from_list(alist: list, keyprop: str, ignore_missing_key=True) -> dict:
    retdict = {}
    for anitem in alist:
        keyval = getvalue(anitem, keyprop)
        if keyval is None:
            if not ignore_missing_key:
                raise KeyError
            else:
                continue
        # Force key to string if it looks like a really big int - fudge for UUID/TXND type keys
        keyval = ikeyval(keyval)
        retdict[keyval] = anitem
    return retdict


def list_find_by_prop(alist: list, propname: str, propval: Union[str,int,float]):
    for item in alist:
        if propname not in item:
            continue
        if item[propname] == propval:
            return item
    return None


def is_number(s:str) -> bool:
    if s is None:
        return False
    try:
        float(s)
        return True
    except ValueError:
        pass

    try:
        import unicodedata
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        pass

    return False

def num_value(s:str) -> Union[float,int]:
    if s is None:
        return None
    if not is_number(s):
        return None
    nval = float(s)
    if nval.is_integer():
        nval = int(nval)
    return nval

def normalize_value(s:str) -> Union[float,int,str]:
    if s is None:
        return None
    if not is_number(s):
        return s
    nval = float(s)
    if nval.is_integer():
        nval = int(nval)
    return nval





from .objutils import ikeyval
from .objutils import dict_from_list
from .objutils import dict_from_object
from .objutils import list_find_by_prop
from .objutils import objlist_from_dictlist
from .objutils import dynobj_from_dict
from .objutils import fixedobj_from_dict
from .objutils import getvalue
from .mapmaker import object_convert_with_map
from .mapmaker import object_empty_with_map


#
# fileparse - A set of utilities for parsing files.
#

import sys
import io
from typing import Callable, Union, Tuple, List
from urllib import parse
import json
import csv
from . import FileSet

def dict_flatten_values(adict: dict) -> dict:
    """

    :param adict:
    :return:
    """
    newdict = {}
    for key, vals in adict.items():
        # Force key to string
        ikey = str(key)
        newdict[ikey] = vals[0]
    return newdict


def dict_from_kvpfile(filepath: str) -> dict:
    """

    :type filepath:
    :param filepath:
    :return: 
    """
    kvps = {}
    kvpfile = open(filepath, "r")
    for aline in kvpfile:
        bline = str.strip(aline)
        if bline != "":
            if "=" in bline:
                kvpair = bline.split("=",2)
                if len(kvpair) == 2:
                    # Force key to string
                    key = str(kvpair[0])
                    kvps[key] = kvpair[1]
    if len(kvps) == 0:
        print("Line splitting failed on {}\n".format(filepath))
        sys.exit(40)
    kvpfile.close()
    return kvps


def dict_from_urlfile(filepath: str) -> dict:
    """

    :param filepath: 
    :return: 
    """
    kvps = {}
    urlfile = open(filepath, "r")
    for aline in urlfile:
        linepairs = parse.parse_qs(aline, True)
        kvps.update(linepairs)
    urlfile.close()
    return dict_flatten_values(kvps)


def dicts_from_files(afileset: FileSet, keyprop: str, filetype="kvp") -> Tuple[dict, list]:
    """

    :param afileset:
    :param keyprop:
    :param filetype:
    :return:
    """
    if "kvp" == filetype:
        ffunc = dict_from_kvpfile
    elif "url" == filetype:
        ffunc = dict_from_urlfile
    else:
        print("Unknown file type {}\n".format(filetype))
        raise NotImplementedError
    dicts = {}
    nokeys = []
    for path in afileset:
        kvpset = ffunc(path)
        if keyprop in kvpset:
            # Force keys to string so the dict is sortable
            # Some PP txn IDs will be all digits so will be int/floats by default
            dkey = str(kvpset[keyprop])
            dicts[dkey] = kvpset
        else:
            nokeys.append(kvpset)
    return  (dicts, nokeys)


def data_from_json_file(filepath: str) -> Union[dict, list, object]:
    """

    :type filepath:
    :param filepath:
    :return:
    """
    kvpfile = open(filepath, "r")
    anobj = json.load(kvpfile)
    kvpfile.close()
    if len(anobj) == 0:
        print("Load JSON failed on {}\n".format(filepath))
        sys.exit(40)
    return anobj


def dictlist_from_csv_stream(csvio) -> list:
    retrows = []
    #csvdialect = csv.Sniffer().sniff(csvio.read(1024))
    #csvio.seek(0)
    reader = csv.DictReader(csvio, fieldnames=None, dialect='excel')
    for row in reader:
        retrows.append(row)
    return retrows


def dictlist_from_csv_str(csvbuf: str) -> list:
    retrows = []
    with io.StringIO(csvbuf) as csvfile:
        retrows = dictlist_from_csv_stream(csvfile)
    return retrows


def dictlist_from_csv_file(path: str) -> list:
    retrows = []
    with open(path, newline='') as csvfile:
        retrows = dictlist_from_csv_stream(csvfile)
    return retrows


def column_from_csv_str(csvbuf: str, colnum: int) -> list:
    retvals = []
    with io.StringIO(csvbuf) as csvfile:
        fieldreader = csv.reader(csvfile)
        for row in fieldreader:
            if colnum < len(row):
                retvals.append(row[colnum])
    return retvals


def column_from_csv_file(path: str, colnum: int) -> list:
    retrows = []
    with open(path, newline='') as csvfile:
        retrows = column_from_csv_str(csvfile, colnum)
    return retrows


def data_to_json_file(filepath: str, odata: Union[dict, List[dict]]):
    """
    Write data to a JSON file
    :type filepath:
    :param filepath:
    :return:
    """
    jsonfile = open(filepath, "w")
    json.dump(odata, jsonfile, sort_keys=True, indent=4)
    jsonfile.close()



#
# datafetch - A set of utilities for fetching data from a database.
#

import sys
import mysql.connector
from typing import Callable, Union, Tuple

class DataFetch:
    def __init__(self):
        self.clear()

    def clear(self):
        return

    def db_init(self, database, user, password, host="127.0.0.1", port=3306, dbtype="mysql"):
        self.db_db = database
        self.db_user = user
        self.db_pw = password
        self.db_host = host
        self.db_port = port
        self.db_type = dbtype
        self.db_set = True

    def db_connect(self, autocommit=True):
        if not self.db_set:
            raise ConnectionError("Connect without DB parameters set")
        if self.db_type == "mysql":
            self.db_conn = mysql.connector.connect(host=self.db_host,
                                                   user=self.db_user,
                                                   passwd=self.db_pw,
                                                   database=self.db_db,
                                                   port=self.db_port,
                                                   autocommit=autocommit)

    def dictlist_query(self, query: str) -> list:
        """
        :query SQL Query String:
        :param filepath:
        :return:
        """
        retlist = []
        db_cur = self.db_conn.cursor(dictionary=True)
        try:
            db_cur.execute(query)
            db_result = db_cur.fetchall()
        except mysql.connector.ProgrammingError as pe:
            print("ERROR: query [{}]".format(query))
            raise pe
        for row in db_result:
            retlist.append(row)
        return retlist


